from functools import total_ordering
from math import ceil
from typing import Mapping
from warnings import warn

from .data import *


class Enum(Converter):
    """
    Base class for enum types

    This implementation is used over Python's builtin solutions to permit interface with the `Converter` system.
    """

    _T = bytes

    _all = []

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `bytes`, returning the first byte

        :param data: The raw bytes to convert
        :return: The first byte of `data`
        """

        return bytes(data[0:1])

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        """
        Converts `bytes` -> `bytes`, enforcing that the input is a recognized enum value

        :param value: The value to convert
        :return: The byte in `value`, unchanged
        """

        if value not in cls._all:
            warn(f"{value} is not recognized.",
                 BytesWarning)

        return value

    @classmethod
    def get_name(cls, value: _T) -> str:
        """
        Finds the first name in this enum with a given value

        :param value: The value to find
        :return: A name in this enum
        """

        return next(filter(lambda attr: not attr.startswith("_") and getattr(cls, attr) == value, dir(cls)), None)


@total_ordering
class Flags(Converter, dict, Mapping[int, int]):
    """
    Base class for flag types

    Flags are bitfields in a byte that are set or cleared using dict update notation.
    """

    _T = 'Flags'

    def __init__(self, bitsets: Mapping[int, int] | 'Flags' = None, *, width: int = 8):
        """
        Create an empty `Flags` instance with a given initial state and width

        :param bitsets: The initial state of these flags
        :param width: The number of bitfields used for these flags (defaults to `8`)
        """

        if bitsets is None:
            bitsets = {bit: 0 for bit in range(width)}

        else:
            bitsets = {bit: 0 for bit in range(ceil((max(bitsets.keys(), default=0) + 1) / 8) * 8)} | bitsets

        super().__init__({bit: value % 2 for bit, value in bitsets.items()})

    def __gt__(self, other) -> bool:
        return int(self) > int(other)

    def __int__(self) -> int:
        return int(str(self), 2)

    def __str__(self) -> str:
        return ''.join([str(bit) for bit in self.values()][::-1])

    def __contains__(self, bitsets: Mapping[int, int] | 'Flags') -> bool:
        return all(self[bit] == int(bool(bitsets[bit])) for bit in bitsets)

    has = __contains__

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `Flags`, splitting the byte into the corresponding bitfields

        :param data: The raw bytes to convert
        :return: A `Flags` instance with bitfields given by `data`
        """

        return cls({bit: int(value) for bit, value in enumerate(f"{int.from_bytes(data, 'little'):b}"[::-1])})

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        """
        Converts `Flags` -> `bytes`, packing the bitfields into the appropriate number of bytes

        :param value: The value to convert
        :return: The byte string composed of the bitfields in these flags
        """

        return int.to_bytes(int(value), len(value) // 8, 'little')


__all__ = ["Enum", "Flags"]



import copy
import inspect

from math import ceil
from typing import Callable, TypeVar
from warnings import warn


_T = TypeVar('_T')


class Converter:
    """
    Abstract base class for data section converters
    """

    _T = _T

    @classmethod
    def get(cls, data: bytes, *, instance=None) -> _T:
        """
        Converts `bytes` -> `_T`

        :param data: The raw bytes to convert
        :param instance: The instance which contains the data section
        :return: An instance of `_T`
        """

        raise NotImplementedError

    @classmethod
    def set(cls, value: _T, *, instance=None, length: int = None, current: bytes = None) -> bytes:
        """
        Converts  `_T` -> `bytes`

        :param value: The value to convert
        :param instance: The instance which contains the data section
        :param length: The length of the data section
        :param current: The current value of the data section
        :return: A string of bytes
        """

        raise NotImplementedError


class Bytes(Converter):
    """
    No-op converter for data sections best interpreted as raw bytes
    """

    _T = bytes

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `bytes` (no-op)

        :param data: The raw bytes to convert
        :return: The bytes in `data`, unchanged
        """

        return bytes(data)

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        """
        Converts `bytes` -> `bytes` (no-op)

        :param value: The value to convert
        :return: The bytes in `value`, unchanged
        """

        return bytes(value)


class Data(Bytes):
    _T = bytes

    @classmethod
    def set(cls, value: _T, *, instance=None, **kwargs) -> bytes:
        """
        Converts `bytes` -> `bytes` and updates metadata fields

        Certain metadata fields are updated automatically based on the entry's data.
        The following are set by this converter:
            - Version
            - Length (for sized data)

        :param value: The value to convert
        :param instance: The instance which contains the data section
        :return: The bytes in `value`, unchanged
        """

        instance.version = instance.get_version(value)
        return super().set(value)


class SizedData(Data):
    _T = bytes

    @classmethod
    def set(cls, value: _T, *, instance=None, **kwargs) -> bytes:
        instance.length = len(value)
        return super().set(value, instance=instance)


class Boolean(Converter):
    """
    Converter for data sections best interpreted as boolean flags

    The data section is expected to have length one.
    """

    _T = bool

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `bool`, where any nonzero value is truthy

        :param data: The raw bytes to convert
        :return: Whether `data` is nonzero
        """

        return data != b'\x00'

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        """
        Converts `bool` -> `bytes`, where `b'\x80'` is truthy and `b'\x00'` is falsy

        :param value: The value to convert
        :return: `b'\x80'` if `value` is truthy else `b'\x00'`
        """

        return b'\x80' if value else b'\x00'


class Integer(Converter):
    """
    Converter for data sections best interpreted as integers

    Integers are always little-endian, unsigned, and at most two bytes.
    """

    _T = int

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `int`

        :param data: The raw bytes to convert
        :return: The little-endian integer given by `data`
        """

        return int.from_bytes(data, 'little')

    @classmethod
    def set(cls, value: _T, *, length: int = None, **kwargs) -> bytes:
        """
        Converts `int` -> `bytes`

        For implementation reasons, the output of this converter is always two bytes wide

        :param value: The value to convert
        :param length: The length of the data section
        :return: The little-endian representation of `value`
        """

        return int.to_bytes(value, length if length is not None else 2, 'little')


class String(Converter):
    """
    Converter for data sections best interpreted as strings

    Strings are encoded in utf-8.
    """

    _T = str

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `str`

        :param data: The raw bytes to convert
        :return: The utf-8 decoding of `data` with trailing null bytes removed
        """

        return data.decode('utf8').rstrip('\0')

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        """
        Converts `str` -> `bytes`

        :param value: The value to convert
        :return: The utf-8 encoding of `value`
        """

        return value.encode('utf8')


class Bits:
    """
    Sliceable converter to extract and package a slice of bits within a byte

    Use like `Bits[start:end:step]` to view a slice of a byte.
    If the slice is empty, the entire byte will be targeted.
    """

    def __class_getitem__(cls, item: slice):
        indices = range(*item.indices(8))

        class BitSlice(Converter):
            """
            Converter to extract and package a slice of bits within a byte

            The data section is expected to have length one.
            """

            _T = int

            mask = sum(1 << i for i in indices)

            @classmethod
            def get(cls, data: bytes, **kwargs) -> _T:
                """
                Converts `bytes` -> `int` by concatenating bits in a slice

                :param data: The raw bytes to convert
                :return: The sliced bits in `data` joined without gaps as an integer
                """

                value = ""
                for index, bit in enumerate(f"{data[0]:08b}"[::-1]):
                    if index in indices:
                        value += bit

                return int(value[::-1], 2)

            @classmethod
            def set(cls, value: _T, *, current: bytes = None, **kwargs) -> bytes:
                """
                Converts `int` -> `bytes` by setting bits in a slice

                :param value: The value to convert
                :param current: The current value of the data section
                :return: The bytes in `value` fit into the section
                """

                data = 0
                bits = iter(f"{value % 256:08b}"[::-1])

                for index in range(8):
                    if index in indices:
                        data += int(next(bits)) << index

                    else:
                        data += current[0] & (1 << index)

                return bytes([data])

        return BitSlice


class Section:
    """
    Data section class which handles conversion between bytes and appropriate data types

    A data section is given by its length and type converter.
    Their primary function is to permit the user to read and write data sections as their natural data types.

    The sections are stored in the `raw` attribute of their class, which is an instance of a `Raw` container.
    A priori, a data section does not correspond to any one part of a var file.
    Individual sections are instead stitched together to yield a var file's contents via `raw.bytes()`.

    Distinct data sections are entirely independent, which is useful for sections which may vary in length.
    To construct sections which point to a portion of another section, see the `View` class.

    Data sections can be declared by decorating methods:
    ```py
    @Section(length, Converter)
    def data_section(self) -> _T:
        ...
    ```

    An optional second parameter can be passed, wherein the method is used as a pre-converter before `Converter.set`.
    """

    def __init__(self, length: int = None, converter: type[Converter] = None):
        """
        Define a new data section given a length and type converter

        :param length: The length of the section (defaults to `None`, i.e. unbounded)
        :param converter: The type converter for the section (defaults to `Bytes`)
        """

        self._converter = converter or Bytes
        self._get, self._set = self._converter.get, self._converter.set
        self._length = length

    def __copy__(self) -> 'Section':
        cls = self.__class__
        new = cls.__new__(cls)
        new.__dict__.update(self.__dict__)
        return new

    def __deepcopy__(self, memo) -> 'Section':
        cls = self.__class__
        new = cls.__new__(cls)
        memo[id(self)] = new

        for k, v in self.__dict__.items():
            setattr(new, k, copy.deepcopy(v, memo))

        return new

    def __set_name__(self, owner, name: str):
        self._name = name

    def __get__(self, instance, owner: type = None) -> _T:
        if instance is None:
            return self

        return self._get(self._get_raw(instance), instance=instance)

    def __set__(self, instance, value: _T):
        setattr(instance.raw, self._name, self._set_raw(instance, value))

    def __call__(self, func: Callable) -> 'Section':
        new = copy.copy(self)
        new.__doc__ = func.__doc__

        signature = inspect.signature(func)
        match len(signature.parameters):
            case 1: pass
            case 2: new._set = lambda value, _set=new._set, **kwargs: _set(func(self, value), **kwargs)
            case _: raise TypeError("Section and View function definitions can only take 1 or 2 parameters.")

        return new

    def _get_raw(self, instance) -> bytes:
        return getattr(instance.raw, self._name, None)

    def _set_raw(self, instance, value: _T) -> _T:
        value = self._set(value, instance=instance, length=self._length, current=self._get_raw(instance))

        if self._length is not None:
            if len(value) > self._length:
                warn(f"Value {value} is too wide for this buffer; truncating to {value[:self._length]}.",
                     BytesWarning)
                value = value[:self._length]

            value = value.ljust(self._length, b'\x00')

        return value

    @property
    def name(self) -> str:
        """
        :return: The `Raw` class attribute name that this section stores to
        """

        return self._name

    @property
    def length(self) -> int | None:
        """
        :return: The length of this section
        """

        return self._length


class View(Section):
    """
    Data view class which handles conversion between portions of data sections and appropriate data types

    A data view is given by its target data section, type converter, and indices within the target.
    Indices must be contiguous and forward-facing.

    Data views are effectively pointers to the data sections they view, converting data in and out as specified.
    Note that data views cannot target other data views; this is done for implementation simplicity.

    Data views can be declared by decorating methods:
    ```py
    @View(section[slice], Converter)
    def data_view(self) -> _T:
        ...
    ```

    An optional second parameter can be passed, wherein the method is used as a pre-converter before `Converter.set`.
    """

    def __init__(self, target: Section, converter: type[Converter], indices: slice = slice(None)):
        """
        Define a new data view given a data section to watch, a type converter, and the portion of the section to view

        :param target: The data section to view
        :param converter: The type converter for the view (defaults to `Bytes`)
        :param indices: The slice of the data section to view (defaults to the entire section)
        """

        super().__init__(None, converter)

        self._target = target
        self._indices = indices

        if self._target.length is None:
            self._length = max(ceil(((self._indices.stop or 0) - (self._indices.start or 0))
                                    // (self._indices.step or 1)), 0)

            if (self._indices.step or 1) > 0:
                if (self._indices.start or 0) >= 0 and (self._indices.stop is None or self._indices.stop < 0):
                    self._length = None

            else:
                if (self._indices.stop or 0) >= 0 and (self._indices.start is None or self._indices.start < 0):
                    self._length = None

        else:
            self._length = len(range(*self._indices.indices(self._target.length)))

    def __set__(self, instance, value: _T):
        getattr(instance.raw, self._target.name)[self._indices] = self._set_raw(instance, value)

    def __getitem__(self, indices: slice) -> 'View':
        return self.__class__(self._target, self._converter, indices)

    def __index__(self) -> slice:
        return self.indices

    def _get_raw(self, instance) -> bytes:
        return getattr(instance.raw, self._target.name)[self._indices]

    @property
    def target(self) -> 'Section':
        return self._target

    @property
    def indices(self) -> slice:
        return self._indices


class Dock:
    """
    Base class to inherit to implement the loader system
    """

    loaders = {}

    def load(self, data):
        """
        Loads data into an instance by delegating to `Loader` methods based on the input's type

        :param data: Any type which the instance might accept
        """
        for loader_types, loader in self.loaders.items():
            if any(isinstance(data, loader_type) for loader_type in loader_types):
                loader(self, data)
                return

        raise TypeError(f"could not find valid loader for type {type(data)}")


class Loader:
    """
    Function decorator to identify methods as data loaders for `Dock` instances

    Specify the loader's accepted type(s) using brackets:
    ```py
    @Loader[int]
    def load_int(self, data: int):
        ...
    ```
    """

    types = ()

    def __init__(self, func):
        self._func = func

    def __call__(self, *args, **kwargs):
        pass

    def __class_getitem__(cls, item: tuple[type, ...] | type) -> type:
        try:
            return type("Loader", (Loader,), {"types": tuple(item)})

        except TypeError:
            return type("Loader", (Loader,), {"types": (item,)})

    def __set_name__(self, owner, name: str):
        owner.loaders = owner.loaders | {self.types: self._func}
        setattr(owner, name, self._func)


__all__ = ["Section", "View", "Dock", "Loader",
           "Converter", "Bytes", "SizedData", "Boolean", "Integer", "String", "Bits"]



from io import BytesIO
from typing import BinaryIO, ByteString, Iterator, Type
from warnings import warn

from tivars.models import *
from tivars.tokenizer import TokenizedString
from .data import *


class TIHeader:
    """
    Parser for var file headers

    All var files require a header which includes a number of magic bytes, data lengths, and a customizable comment.
    """

    class Raw:
        """
        Raw bytes container for `TIHeader`

        Any class with a distinct byte format requires its own `Raw` class to contain its data sections.
        Each data section must have a corresponding slot in `Raw` in order to use `Converter` classes.

        The `Raw` class must also contain a `bytes()` method specifying the order and visibility of the data sections.
        Additional methods can also be included, but should be callable from the outer class.
        """

        __slots__ = "magic", "extra", "product_id", "comment"

        def bytes(self) -> bytes:
            """
            :return: The bytes contained in this header
            """

            return self.magic + self.extra + self.product_id + self.comment

    def __init__(self, model: TIModel = None, *,
                 magic: str = None, extra: bytes = b'\x1a\x0a', product_id: int = None,
                 comment: str = "Created with tivars_lib_py v0.7.0"):
        """
        Create an empty header which targets a specified model

        :param model: A minimum `TIModel` to target (defaults to `TI_83P`)
        :param magic: File magic at the start of the header (default to the model's magic)
        :param extra: Extra export bytes for the header (defaults to `0x1a0a`)
        :param product_id: The targeted model's product ID (defaults to `0x00`)
        :param comment: A comment to include in the header (defaults to a simple lib message)
        """

        self.raw = self.Raw()

        model = model or TI_83P

        self.magic = magic or model.magic
        self.extra = extra
        self.product_id = product_id if product_id is not None else model.product_id
        self.comment = comment

    def __bytes__(self) -> bytes:
        return self.bytes()

    def __copy__(self) -> 'TIHeader':
        new = TIHeader()
        new.load_bytes(self.bytes())
        return new

    def __eq__(self, other: 'TIHeader') -> bool:
        try:
            return self.__class__ == other.__class__ and self.bytes() == other.bytes()

        except AttributeError:
            return False

    def __or__(self, other: list['TIEntry']):
        new = other[0].export(header=self, name=other[0].name, model=self.targets().pop())

        for entry in other[1:]:
            new.add_entry(entry)

        return new

    def __len__(self) -> int:
        return 53

    @Section(8, String)
    def magic(self) -> str:
        """
        The file magic for the var
        """

    @Section(2)
    def extra(self) -> bytes:
        """
        Extra export bytes for the var

        The exact meaning and interpretation of these bytes is not yet determined.
        These bytes are set by different export tools and can often be "incorrect" without causing issues.
        """

    @Section(1, Bits[:])
    def product_id(self) -> int:
        """
        The product ID for the var

        While used to identify the model the var was created on, it has no actual functional ramifications.
        Furthermore, it does not constitute a 1-to-1 mapping to distinct models.
        """

    @Section(42, String)
    def comment(self) -> str:
        """
        The comment attached to the var
        """

    def targets(self) -> set[TIModel]:
        """
        Determine which model(s) this header can target

        The header contains no reference to a model to target, which permits sharing across models where possible.
        This method derives a set of valid model from the header's file magic and product ID.

        If the header contains malformed magic, an error will be raised.
        If the header contains a malformed product ID, it will be ignored.

        :return: A set of models that this header can target
        """

        models = {m for m in TIModel.MODELS if m.magic == self.magic}

        if self.product_id != b'\x00':
            if filtered := {m for m in models if m.product_id == self.product_id}:
                return filtered

        if not models:
            raise ValueError(f"file magic '{self.magic}' not recognized")

        return models

    def load_bytes(self, data: bytes | BytesIO):
        """
        Loads a byte string or bytestream into the header

        :param data: The source bytes
        """

        try:
            data = BytesIO(data.read())

        except AttributeError:
            data = BytesIO(data)

        # Read magic
        self.raw.magic = data.read(8)

        # Read export bytes
        self.raw.extra = data.read(2)

        # Read product ID
        self.raw.product_id = data.read(1)

        # Read comment
        self.raw.comment = data.read(42)

    def bytes(self) -> bytes:
        """
        :return: The byte string corresponding to this header
        """

        return self.raw.bytes()

    def load_from_file(self, file: BinaryIO):
        """
        Loads this header from a file given a file pointer

        :param file: A binary file to read from
        """

        self.load_bytes(file.read(len(self)))

    def open(self, filename: str):
        """
        Loads this header from a file given a filename

        :param filename: A filename to open
        """

        with open(filename, 'rb') as file:
            self.load_bytes(file.read())


class TIEntry(Dock, Converter):
    """
    Base class for all var entries

    A var file is made of one or more entries, each of which contain the data of the familiar var types.
    """

    _T = 'TIEntry'

    flash_only = False
    """
    Whether this entry only supports flash chips
    """

    extensions = {None: "8xg"}
    """
    The file extension used for this entry per-model
    """

    versions = [0x00]
    """
    The possible versions of this entry
    """

    base_meta_length = 11
    flash_meta_length = 13

    min_data_length = 0
    """
    The minimum length of this entry's data
    
    If an entry's data is fixed in size, this value is necessarily the length of the data
    """

    leading_bytes = b''
    """
    Bytes that always occur at the start of this entry's data
    """

    _type_id = None
    _type_ids = {}

    class Raw:
        """
        Raw bytes container for `TIEntry`

        Any class with a distinct byte format requires its own `Raw` class to contain its data sections.
        Each data section must have a corresponding slot in `Raw` in order to use `Converter` classes.

        The `Raw` class must also contain a `bytes()` method specifying the order and visibility of the data sections.
        Additional methods can also be included, but should be callable from the outer class.

        Most entry types do not require a new `Raw` class since only the entry's data changes between types.
        """

        __slots__ = "meta_length", "type_id", "name", "version", "archived", "calc_data"

        @property
        def data_length(self) -> bytes:
            """
            :return: The length of this entry's data portion
            """

            return int.to_bytes(len(self.calc_data), 2, 'little')

        @property
        def flash_bytes(self) -> bytes:
            """
            :return: The flash bytes of this entry if they exist
            """

            return (self.version + self.archived)[
                   :int.from_bytes(self.meta_length, 'little') - TIEntry.base_meta_length]

        @property
        def meta(self) -> bytes:
            """
            :return: The meta section of this entry
            """

            return self.bytes()[2:int.from_bytes(self.meta_length, 'little') + 2]

        def bytes(self) -> bytes:
            """
            :return: The bytes contained in this entry
            """

            return self.meta_length + self.data_length + \
                self.type_id + self.name + self.version + self.archived + \
                self.data_length + self.calc_data

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "UNNAMED",
                 version: bytes = None, archived: bool = None,
                 data: bytes = None):
        """
        Creates an empty entry with specified meta and data values

        :param init: Values to initialize this entry's data (defaults to `None`)
        :param for_flash: Whether this entry supports flash chips (defaults to `True`)
        :param name: The name of this entry (defaults to a valid default name)
        :param version: This entry's version (defaults to `None`; not supported if `for_flash == False`)
        :param archived: Whether this entry is archived (defaults to `False`; not supported if `for_flash == False`)
        :param data: This entry's data (defaults to empty)
        """

        self.raw = self.Raw()

        self.meta_length = TIEntry.flash_meta_length if for_flash else TIEntry.base_meta_length
        self.type_id = self._type_id if self._type_id is not None else 0xFF
        self.name = name
        self.archived = archived or False

        if not for_flash:
            if version is not None or archived is not None:
                warn("Models without flash chips do not support versioning or archiving.",
                     UserWarning)

            if self.flash_only:
                warn(f"{type(self)} entries are not compatible with flashless chips.",
                     UserWarning)

        self.clear()
        if data:
            self.data = bytearray(data)
            self.coerce()

        elif init is not None:
            try:
                self.load_bytes(init.bytes())
            except AttributeError:
                self.load(init)

        self.version = self.get_version()

    def __bool__(self) -> bool:
        return not self.is_empty

    def __bytes__(self) -> bytes:
        return self.bytes()

    def __copy__(self) -> 'TIEntry':
        new = self.__class__()
        new.load_bytes(self.bytes())
        return new

    def __eq__(self, other: 'TIEntry') -> bool:
        try:
            return self.__class__ == other.__class__ and self.bytes() == other.bytes()

        except AttributeError:
            return False

    def __format__(self, format_spec: str) -> str:
        raise TypeError(f"unsupported format string passed to {type(self)}.__format__")

    def __init_subclass__(cls, /, register=False, **kwargs):
        super().__init_subclass__(**kwargs)

        if register:
            TIEntry.register(cls)

    def __iter__(self) -> Iterator:
        raise NotImplementedError

    def __len__(self) -> int:
        return 2 + self.meta_length + 2 + self.data_length

    def __str__(self) -> str:
        return self.string()

    @Section(2, Integer)
    def meta_length(self) -> int:
        """
        The length of the meta section of the entry

        The possible meta lengths are 11 (without flash) or 13 (with flash).
        """

    @property
    def data_length(self) -> int:
        """
        The length of the data section of the entry
        """

        return len(self.calc_data)

    @Section(1, Bits[:])
    def type_id(self) -> int:
        """
        The type ID of the entry

        The type determines how the contents of the data section of the entry are interpreted.
        """

    @Section(8, TokenizedString)
    def name(self) -> str:
        """
        The name of the entry

        Interpretation as text depends on the entry type; see individual types for details.
        """

    @Section(1, Bits[:])
    def version(self) -> bytes:
        """
        The version number of the entry

        The version is used to determine model compatibility where necessary.
        Only flash files support this entry, and it thus not present if `meta_length` <= 11.
        """

    @Section(1, Boolean)
    def archived(self) -> bool:
        """
        Whether the entry is archived

        Only flash files support this entry, and it thus not present if `meta_length` <= 11.
        """

    @Section()
    def calc_data(self) -> bytearray:
        """
        The data section of the entry which is loaded on-calc
        """

    @View(calc_data, Bytes)[:]
    def data(self) -> bytearray:
        """
        The entry's user data
        """

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `TIEntry`

        :param data: The raw bytes to convert
        :return: A `TIEntry` instance with data equal to `data`
        """

        return cls(data=data)

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        """
        Converts `TIEntry` -> `bytes`

        :param value: The value to convert
        :return: The data of `value`
        """

        return value.calc_data

    @property
    def flash_bytes(self) -> bytes:
        """
        :return: The flash bytes of this entry if they exist
        """

        return (self.raw.version + self.raw.archived)[:self.meta_length - TIEntry.base_meta_length]

    @property
    def is_empty(self) -> bool:
        """
        :return: Whether this entry's data is empty
        """

        return self.data_length == 0

    @property
    def meta(self) -> bytes:
        """
        :return: The meta section of this entry
        """

        return self.raw.data_length + self.raw.type_id + self.raw.name + self.raw.version + self.raw.archived

    @classmethod
    def get_type(cls, type_id: int) -> Type['TIEntry']:
        """
        Gets the subclass corresponding to a type ID if one is registered

        :param type_id: The type ID to search by
        :return: A subclass of `TIEntry` with corresponding type ID or `None`
        """

        return cls._type_ids.get(type_id, None)

    @staticmethod
    def next_entry_length(stream: BinaryIO) -> int:
        """
        Helper function to determine the length of the next entry in a bytestream

        :param stream: A bytestream
        :return: The length of the next entry in the bytestream
        """

        meta_length = int.from_bytes(stream.read(2), 'little')
        data_length = int.from_bytes(stream.read(2), 'little')
        stream.seek(-4, 1)

        return 2 + meta_length + 2 + data_length

    @classmethod
    def register(cls, var_type: type['TIEntry']):
        """
        Registers a subtype with this class for coercion

        :param var_type: The `TIEntry` subtype to register
        """

        cls._type_ids[var_type._type_id] = var_type

    def archive(self):
        """
        Archives this entry (if supported)
        """

        if self.flash_bytes:
            self.archived = True
        else:
            raise TypeError("entry does not support archiving.")

    def clear(self):
        """
        Clears this entry's data
        """

        self.raw.calc_data = bytearray(self.leading_bytes)
        self.raw.calc_data.extend(bytearray(self.min_data_length - self.data_length))

    def get_min_os(self, data: bytes = None) -> OsVersion:
        """
        Determines the minimum OS that supports this entry's data

        :param data: The data to find the minimum support for (defaults to this entry's data)
        :return: The minimum `OsVersion` this entry supports
        """

        return OsVersions.INITIAL

    def get_version(self, data: bytes = None) -> int:
        """
        Determines the version byte corresponding to given data for this entry type

        Entries which could contain non-backwards compatible data are assigned a version byte.
        If an entry's version exceeds the "version" of a calculator, transfer to the calculator will fail.

        :param data: The data to find the version of (defaults to this entry's data)
        :return: The version byte for `data`
        """

        return self.versions[0]

    def supported_by(self, model: TIModel) -> bool:
        """
        Determines whether a given model can support this entry

        :param model: The model to check support for
        :return: Whether `model` supports this entry
        """

        return self.get_min_os() < model.OS("latest")

    def unarchive(self):
        """
        Unarchives this entry (if supported)
        """

        if self.flash_bytes:
            self.archived = False
        else:
            raise TypeError("entry does not support archiving.")

    @Loader[ByteString, BytesIO]
    def load_bytes(self, data: bytes | BytesIO):
        """
        Loads a byte string or bytestream into this entry

        :param data: The bytes to load
        """

        try:
            data = BytesIO(data.read())

        except AttributeError:
            data = BytesIO(data)

        # Read meta length
        self.raw.meta_length = data.read(2)

        # Read data length
        data_length = data.read(2)

        # Read and check type ID
        self.raw.type_id = data.read(1)

        if self._type_id is not None and self.type_id != self._type_id:
            if subclass := TIEntry.get_type(self.type_id):
                warn(f"The entry type is incorrect (expected {type(self)}, got {subclass}).",
                     BytesWarning)

            else:
                warn(f"The entry type is incorrect (expected {type(self)}, got an unknown type). "
                     f"Load the var file into a TIVar instance if you don't know the entry type(s).",
                     BytesWarning)

        # Read varname
        self.raw.name = data.read(8)

        # Read flash bytes
        match self.meta_length:
            case TIEntry.flash_meta_length:
                self.raw.version = data.read(1)
                self.raw.archived = data.read(1)

                if self.raw.archived not in b'\x00\x80':
                    warn(f"The archive flag ({self.raw.archived.hex()}) is set to an unexpected value.",
                         BytesWarning)

            case TIEntry.base_meta_length:
                self.raw.version = b'\x00'
                self.raw.archived = b'\x00'

                if self.flash_only:
                    warn(f"{type(self)} vars are not compatible with flashless chips.",
                         BytesWarning)

            case _:
                warn(f"The entry meta length has an unexpected value ({self.meta_length}); "
                     f"attempting to read flash bytes anyway.",
                     BytesWarning)
                self.raw.version = data.read(1)
                self.raw.archived = data.read(1)

                if self.raw.archived not in b'\x00\x80':
                    warn(f"The archive flag is set to an unexpected value.",
                         BytesWarning)

        # Read data and check length
        data_length2 = data.read(2)
        if data_length != data_length2:
            warn(f"The var entry data lengths are mismatched ({data_length} vs. {data_length2}); "
                 f"using {data_length2} to read the data section.",
                 BytesWarning)

        self.raw.calc_data = bytearray(data.read(int.from_bytes(data_length2, 'little')))

        self.coerce()

        if self.versions != [0x00] and self.version not in self.versions:
            warn(f"The version (0x{self.version:02x}) is not recognized.",
                 BytesWarning)

    def bytes(self) -> bytes:
        """
        :return: The byte string corresponding to this entry
        """

        return self.raw.bytes()

    def load_data_section(self, data: BytesIO):
        """
        Loads the data of this entry from a bytestream

        :param data: The source bytes
        """

        self.raw.calc_data = bytearray(data.read(type(self).calc_data.length))

    @Loader[BinaryIO]
    def load_from_file(self, file: BinaryIO, *, offset: int = 0):
        """
        Loads this entry from a file given a file pointer and offset

        :param file: A binary file to read from
        :param offset: The offset of the entry to read
        """

        # Load header
        header = TIHeader()
        header.load_from_file(file)
        file.seek(2, 1)

        # Seek to offset
        while offset:
            file.seek(self.next_entry_length(file), 1)
            offset -= 1

        self.load_bytes(file.read(self.next_entry_length(file)))
        file.seek(2, 1)

    @Loader[str]
    def load_string(self, string: str):
        """
        Loads this entry from a string representation

        :param string: The string to load
        """

        raise NotImplementedError

    def string(self) -> str:
        """
        :return: A string representation of this entry
        """

        raise NotImplementedError

    def open(self, filename: str):
        """
        Loads this header from a file given a filename

        :param filename: A filename to open
        """

        if self._type_id is not None and \
                not any(filename.endswith(extension) for extension in self.extensions.values()):
            warn(f"File extension .{filename.split('.')[-1]} not recognized for var type {type(self)}; "
                 f"attempting to read anyway.")

        with open(filename, 'rb') as file:
            file.seek(55)
            self.load_bytes(file.read(self.next_entry_length(file)))
            file.seek(2, 1)

            if file.read():
                warn("The selected var file contains multiple entries; only the first will be loaded. "
                     "Use load_from_file to select a particular entry, or load the entire file in a TIVar object.",
                     UserWarning)

    def save(self, filename: str = None, *, header: TIHeader = None, model: TIModel = None):
        """
        Save this entry as a var file given a filename and optional header and targeted model

        :param filename: A filename to save to (defaults to the var's name and extension)
        :param header: A `TIHeader` to attach (defaults to an empty header)
        :param model: A `TIModel` to target (defaults to `None`)
        """

        self.export(header=header, model=model).save(filename)

    def export(self, *, name: str = None, header: TIHeader = None, model: TIModel = None) -> 'TIVar':
        """
        Export this entry to a `TIVar` with a specified name, header, and target model

        :param name: The name of the var (defaults to this entry's name)
        :param header: A `TIHeader` to attach (defaults to an empty header)
        :param model: A `TIModel` to target (defaults to `None`)
        """

        var = TIVar(header=header, name=name or self.name, model=model)
        var.add_entry(self)
        return var

    def coerce(self):
        """
        Coerces this entry to a subclass if possible using the entry's type ID

        Valid types must be registered to be considered for coercion.
        """

        if self._type_id is None:
            if subclass := self.get_type(self.type_id):
                self.__class__ = subclass
                self.coerce()

            elif self.type_id != 0xFF:
                warn(f"Type ID 0x{self.type_id:02x} is not recognized; entry will not be coerced to a subclass.",
                     BytesWarning)


class TIVar:
    """
    Container for var files

    A var file is composed of a header and any number of entries (though most have only one).
    """

    def __init__(self, *, name: str = "UNNAMED", header: TIHeader = None, model: TIModel = None):
        """
        Creates an empty var with a specified name, header, and targeted model

        :param name: The name of the var (defaults to `UNNAMED`)
        :param header: A `TIHeader` to attach (defaults to an empty header)
        :param model: A minimum `TIModel` to target (defaults to `None`)
        """

        self._header = header or TIHeader(model)
        self.entries = []

        self.name = name
        self._model = model

        if self._model and self._model not in self._header.targets():
            warn(f"The var's model ({self._model}) is incompatible with the given header.",
                 UserWarning)

    def __bool__(self) -> bool:
        return not self.is_empty

    def __bytes__(self) -> bytes:
        return self.bytes()

    def __copy__(self) -> 'TIVar':
        new = TIVar()
        new.load_bytes(self.bytes())
        return new

    def __eq__(self, other: 'TIVar'):
        try:
            eq = self.__class__ == other.__class__ and len(self.entries) == len(other.entries)
            return eq and all(entry == other_entry for entry, other_entry in zip(self.entries, other.entries))

        except AttributeError:
            return False

    def __len__(self):
        return len(self._header) + self.entry_length + 2

    @property
    def entry_length(self) -> int:
        """
        The total length of the var entries

        This should always be 57 less than the total var size.
        """

        return sum(len(entry) for entry in self.entries)

    @property
    def checksum(self):
        """
        The checksum for the var

        This is equal to the lower 2 bytes of the sum of all bytes in the entries.
        """

        return int.to_bytes(sum(sum(entry.bytes()) for entry in self.entries) & 0xFFFF, 2, 'little')

    @property
    def extension(self) -> str:
        """
        Determines the var's file extension based on its entries and targeted model.

        If there is only one entry, that entry's extension for the target model is used.
        Otherwise, `.8xg` is used.

        :return: The var's file extension
        """

        if len(self.entries) > 1:
            return "8xg"

        try:
            if self._model is None:
                return self.entries[0].extensions[None]

            extension = ""
            for model in reversed(TIModel.MODELS):
                if model in self.entries[0].extensions and model <= self._model:
                    extension = self.entries[0].extensions[self._model]
                    break

            if not extension:
                warn(f"The {self._model} does not support this var type.",
                     UserWarning)

                return self.entries[0].extensions[None]

            return extension

        except IndexError:
            raise ValueError("this var is empty")

    @property
    def filename(self) -> str:
        """
        Determines the var's filename based on its name, entries, and targeted model.

        The filename is the concatenation of the var name and extension (see `TIVar.extension`).

        :return: The var's filename
        """

        return f"{self.name}.{self.extension}"

    @property
    def header(self) -> 'TIHeader':
        """
        :return: This var's header
        """

        return self._header

    @property
    def is_empty(self) -> bool:
        """
        :return: Whether this var contains no entries
        """

        return len(self.entries) == 0

    @property
    def model(self) -> TIModel:
        """
        :return: This var's targeted model
        """

        return self._model

    def add_entry(self, entry: TIEntry = None):
        """
        Adds an entry to this var

        :param entry: A `TIEntry` to add (defaults to an empty entry)
        """

        entry = entry or TIEntry()

        if not self.is_empty:
            if entry.meta_length != self.entries[0].meta_length:
                warn(f"The new entry has a conflicting meta length "
                     f"(expected {self.entries[0].meta_length}, got {entry.meta_length}).",
                     UserWarning)

        self.entries.append(entry)

    def clear(self):
        """
        Removes all entries from this var
        """

        self.entries.clear()

    def supported_by(self, model: TIModel = None) -> bool:
        """
        Determines whether a given model can support this var

        :param model: The model to check support for (defaults to this var's model, if it is set)
        :return: Whether `model` supports this var
        """

        model = model or self._model
        if model is None:
            raise ValueError("no model was passed")

        return model in self._header.targets() and \
            all(entry.get_min_os() < model.OS("latest") for entry in self.entries)

    def load_bytes(self, data: bytes | BytesIO):
        """
        Loads a byte string or bytestream into this var

        :param data: The bytes to load
        """

        try:
            data = BytesIO(data.read())

        except AttributeError:
            data = BytesIO(data)

        # Read header
        self._header.load_bytes(data.read(53))
        entry_length = int.from_bytes(data.read(2), 'little')

        # Read entries
        self.clear()
        while entry_length:
            self.add_entry()

            length = TIEntry.next_entry_length(data)
            self.entries[-1].load_bytes(data.read(length))

            entry_length -= length

        # Read checksum
        checksum = data.read(2)

        # Check model
        if self._model and not self._model <= min(*self._header.targets()):
            warn(f"The loaded var file is incompatible with the {self._model}.",
                 BytesWarning)

        # Check² sum
        if checksum != self.checksum:
            warn(f"The checksum is incorrect (expected {self.checksum}, got {checksum}).",
                 BytesWarning)

    def bytes(self):
        """
        :return: The byte string corresponding to this var
        """

        dump = self._header.bytes()
        dump += int.to_bytes(self.entry_length, 2, 'little')

        for entry in self.entries:
            dump += entry.bytes()

        dump += self.checksum
        return dump

    def load_var_file(self, file: BinaryIO):
        """
        Loads this var from a file given a file pointer

        :param file: A binary file to read from
        """

        self.load_bytes(file.read())

    def open(self, filename: str):
        """
        Loads this var from a file given a filename

        :param filename: A filename to open
        """

        with open(filename, 'rb') as file:
            self.load_bytes(file.read())

    def save(self, filename: str = None):
        """
        Save this var given a filename

        :param filename: A filename to save to (defaults to the var's name and extension)
        """

        if self._model:
            for index, entry in enumerate(self.entries):
                if entry.get_min_os() > self._model.OS("latest"):
                    warn(f"Entry #{index + 1} is not supported by {self._model}.",
                         UserWarning)

        with open(filename or self.filename, 'wb+') as file:
            file.write(self.bytes())


class SizedEntry(TIEntry):
    """
    Base class for all sized entries

    A sized entry is an entry with variable-length data, where the length of the data is stored in the first two bytes.
    This length is two less than the length stored in the `data_length` section(s).
    """

    min_data_length = 2

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Integer)[0:2]
    def length(self) -> int:
        """
        The length of this entry's user data section
        """

    @View(calc_data, SizedData)[2:]
    def data(self) -> bytearray:
        pass

    def clear(self):
        self.raw.calc_data = bytearray([0, 0, *self.leading_bytes])
        self.raw.calc_data.extend(bytearray(self.min_data_length - self.data_length))
        self.length = len(self.leading_bytes) + len(self.data)

    def load_bytes(self, data: ByteString):
        super().load_bytes(data)

        if self.length != (data_length := len(self.leading_bytes) + len(self.data)):
            warn(f"The entry has an unexpected data length (expected {self.length}, got {data_length}).",
                 BytesWarning)

    def load_data_section(self, data: BytesIO):
        data_length = int.from_bytes(length_bytes := data.read(2), 'little')
        self.raw.calc_data = bytearray(length_bytes + data.read(data_length))


__all__ = ["TIHeader", "TIEntry", "TIVar", "SizedEntry"]
